/**
 * An array-based implementation of a binary min-heap that stores ints
 */

import java.util.Arrays;
import java.util.NoSuchElementException;
import java.util.Random;

public class MinIntHeap {

    private int[] heap;
    private int size;

    public MinIntHeap() {
        heap = new int[10];
        size = 0;
    }

    public MinIntHeap(int[] input, int size) {
        heap = input;
        this.size = size;

        // TODO: heapify, see Bailey problem 13.10

        checkHeap();
    }

    public int size() {
        return size;
    }

    public boolean isEmpty() {
        return size == 0;
    }

    private void grow() {
        checkHeap();
        int[] newHeap = new int[heap.length * 2];
        for (int i = 0; i < heap.length; i++) {
            newHeap[i] = heap[i];
        }
        heap = newHeap;
        checkHeap();
    }

    public void add(int val) {
        checkHeap();
        if (size == heap.length - 1) {
            grow();
        }
        size++;
        int insertIndex = percolateUp(size, val);
        heap[insertIndex] = val;
        checkHeap();
    }

    private int percolateUp(int hole, int val) {
        while (hole > 1 && val < heap[hole / 2]) {
            heap[hole] = heap[hole / 2];
            hole /= 2;
        }
        return hole;
    }

    public int removeMin() {
        if (isEmpty()) throw new NoSuchElementException();
        checkHeap();
        int ans = heap[1];
        int hole = percolateDown(1, heap[size]);
        heap[hole] = heap[size];
        size--;
        checkHeap();
        return ans;
    }

    private int percolateDown(int hole, int val) {
        while (hole <= size / 2) { // while hole is not a bottom-row leaf
            // find smaller child
            int left = hole * 2;
            int right = left + 1;
            int target = left;
            if (right < size && heap[right] < heap[left]) {
                target = right;
            }
            // swap hole down the tree
            if (heap[target] < val) {
                heap[hole] = heap[target];
                hole = target;
            } else {
                break;
            }
        }
        return hole;
    }

    private void checkHeap() {
        // TODO: throw a RuntimeException if the current array is not a valid heap
    }

    public static void main(String[] args) {
        // test normal heap construction
        MinIntHeap testHeap = new MinIntHeap();
        Random rng = new Random();
        while (testHeap.size() < 10) {
            testHeap.add(rng.nextInt(100));
        }
        while (testHeap.size() > 0) {
            System.out.print(testHeap.removeMin() + " ");
        }
        System.out.println();
    }

    // public static void main(String[] args) {
    //     // test heapifying heap construction
    //     Random rng = new Random();
    //     int[] input = new int[10];
    //     for (int i = 1; i < input.length; i++) {
    //     input[i] = rng.nextInt(1000);
    //     }

    //     MinIntHeap testHeap = new MinIntHeap(input, input.length - 1);
    //     while (testHeap.size() > 0) {
    //     System.out.print(testHeap.removeMin() + " ");
    //     }
    //     System.out.println();
    // }
}